# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2021-07-15
#


import copy


class SimpleAttributeDict(dict):
    """Works like dict but provides methods to get attribute by '.'.

    such as::
        attrDict = SimpleAttributeDict({'foo': foo, 'bar': bar})
        print(attrDict.foo)
        attrDict.bar = bar2
    """

    def __init__(self, defaults={}):
        super(SimpleAttributeDict, self).__init__(defaults)
        # dict.__init__(self, defaults)

    def __getattr__(self, attrname):
        return self.get(attrname)

    def __setattr__(self, attrname, value):
        self[attrname] = value

    def __deepcopy__(self, memo):
        return SimpleAttributeDict(copy.deepcopy(self.copy(), memo))


class AttributeDict(dict):
    """Works like dict but provides methods to get attribute by '.',
    inherited from EasyDict from Mathieu Leplatre

    such as::
        attrDict = AttributeDict({'foo': foo, 'bar': bar})
        attrDict.foo    # => foo
        attrDict.bar = bar2
    """

    def __init__(self, defaults={}):
        # super(AttributeDict, self).__init__(defaults)

        defaults = defaults or {}
        for k, v in defaults.items():
            setattr(self, k, v)

        # Class attributes
        for k in self.__class__.__dict__.keys():
            if not (k.startswith('__') and k.endswith('__')) and \
               k not in ('update', 'pop'):
                setattr(self, k, getattr(self, k))

    def __setattr__(self, name, value):
        if isinstance(value, (list, tuple)):
            value = [self.__class__(x)
                     if isinstance(x, dict) else x for x in value]
        elif isinstance(value, dict) and not isinstance(value, self.__class__):
            value = self.__class__(value)

        super(AttributeDict, self).__setattr__(name, value)
        super(AttributeDict, self).__setitem__(name, value)

    def __deepcopy__(self, memo):
        return AttributeDict(copy.deepcopy(self.copy(), memo))

    __setitem__ = __setattr__

    def update(self, other=None):
        other = other or dict()
        for k in other:
            setattr(self, k, other[k])

    def pop(self, k, default=None):
        delattr(self, k)
        return super(AttributeDict, self).pop(k, default)

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2021-07-15
#


import os
import yaml
import requests
from .attribute_dict import AttributeDict
from .api import API
from .session import Session
from .response import Response


def make_api_method(api, **kwargs):
    def _api(self, **kwargs):
        return self.request(api, **kwargs)
    return _api


class Zentao:
    """zentao main entry class"""

    def __init__(self, config):
        super(Zentao, self).__init__()

        # load config
        self._load_config(config)

        # add dynamical methods for api
        self._init_apis()

        # initialize session
        self.session = Session(
            username=self.config.username,
            password=self.config.password,
            session_api=self.apis.get("api_getSessionID"),
            login_api=self.apis.get("user_login")
        )

# public
    def request(self, api_name, **kwargs):
        """wrapper for requests.request"""

        self.session.connect()

        # check out params
        params = kwargs.pop("params", {})
        params[self.session.name] = self.session.id

        api = self.apis.get(api_name, kwargs)

        response = requests.request(
            method=api.get("method", "GET"),
            url=api.get("url"),
            params=params
        ).json()

        # return raw data or not
        if kwargs.pop("raw", False):
            return response
        else:
            return Response(response)

# protected
    def _init_apis(self):
        """initialize apis"""

        # load api specs
        self.apis = API(
            base_url=self.config.url,
            version=self.config.version,
            config=self.config.get("spec", None)
        )

        # generate methods by api name
        for api in self.apis.names():
            setattr(Zentao, api, make_api_method(api))

    def _load_config(self, config):
        """load configuration"""

        # load customized config
        if isinstance(config, str) and os.path.exists(config):
            # config file path
            with open(config, mode="r", encoding="utf-8") as f:
                cfg = yaml.full_load(f.read()).get("zentao")
        elif isinstance(config, dict):
            cfg = config

        self.config = AttributeDict({
            "url": "",
            "version": "15",
            "username": "",
            "password": "",
        })
        self.config.update(cfg)

        # check out url
        if not self.config.url.endswith("/"):
            self.config.url += "/"

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2021-08-01
#


class PyZentaoException(Exception):
    """pyzentao base exception class"""

    pass


class AuthenticationError(PyZentaoException):
    """exception for authentication error"""

    pass


class APINameError(PyZentaoException):
    """exception for unknown api name error"""

    def __init__(self, name):
        super(APINameError, self).__init__()

        self.name = name

    def __str__(self):
        """printer"""

        return "Unknown API name '%s'" % self.name

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2021-07-19
#


import os
import yaml
import urllib
from .exceptions import APINameError


class API:
    """zentao api collection"""

    def __init__(
        self,
        base_url,
        version,
        config=None
    ):
        super(API, self).__init__()

        self.base_url = base_url
        self.version = version
        self._load(config)

# public
    def get(self, name, params=None):
        """get api config by api name"""

        spec = self._specs.get(name, None)
        if spec is None:
            raise APINameError(name)

        spec["url"] = self._get_url(name, params)
        return spec

    def names(self):
        """wrapper for get data keys"""

        return self._specs.keys()

# protected
    def _get_url(self, name, params):
        """join root and path to make api url"""

        # combine path by spec and params
        spec = self._specs.get(name)
        paths = [spec["path"]]

        if "params" in spec and isinstance(params, dict):
            # placeholder for parameter
            keep = spec.get("keep", False) or params.pop("keep", False)

            for param in spec["params"]:
                if param in params:
                    paths.append(str(params[param]))
                elif keep:
                    paths.append("")

        return "%s.json" % urllib.parse.urljoin(
            self.base_url, "-".join(paths)
        )

    def _load(self, config):
        """load spec data"""

        self._specs = {}

        if config is None:
            self._load_default()
        else:
            if config.get("merge", True):
                self._load_default()

            self._load_path(config.get("path"))

    def _load_default(self):
        """load default spec"""

        self._load_path(
            os.path.join(
                os.path.dirname(__file__),
                "specs",
                "v%s" % str(self.version)
            )
        )

    def _load_path(self, path):
        """load spec from the specified path"""

        if os.path.isdir(path):
            # path is a directory
            for file_name in os.listdir(path):
                self._load_file(os.path.join(path, file_name))
        else:
            # path is a file
            self._load_file(path)

    def _load_file(self, path):
        """load spec from the specified file path"""

        if os.path.exists(path) and path.endswith(".yml"):
            with open(path, mode="r", encoding="utf-8") as f:
                self._specs.update(yaml.full_load(f.read()))


# end
